"""
Created on 2025-02-26

@author: wf
"""

import tarfile
from pathlib import Path


class Tarball:
    """
    tarball support
    """

    @staticmethod
    def read_from_tar(tarball_path: Path, filename: str) -> bytes:
        """
        Reads a file directly from a tarball.

        Args:
            tarball_path (Path): Path to the tar archive.
            filename (str): Name of the file inside the archive.

        Returns:
            bytes: The file contents.
        """
        with tarfile.open(tarball_path, "r") as tar:
            member = tar.getmember(filename)
            with tar.extractfile(member) as file:
                bytes_read = file.read()
                return bytes_read



"""
Created on 2025-02-25


@author: wf
"""

import logging
import mimetypes
import traceback
from pathlib import Path
from typing import Optional, Tuple

from fastapi import FastAPI, HTTPException
from fastapi.responses import FileResponse, HTMLResponse, Response
from fastapi.staticfiles import StaticFiles

from djvuviewer.djvu_config import DjVuConfig
from djvuviewer.djvu_core import DjVuFile, DjVuViewPage
from djvuviewer.image_convert import ImageConverter
from djvuviewer.tarball import Tarball


class DjVuViewer:
    """
    Handles loading DjVu metadata, serving static content, and browsing via DB or API.
    """

    _static_mounted = False

    def __init__(self, app: FastAPI, config: DjVuConfig):
        """
        Initialize the DjVu viewer.

        Args:
            app: FastAPI application instance
            config:DjVu configuration
        """
        self.config = config
        self.url_prefix = self.config.url_prefix.rstrip("/")
        self.app = app

        if not DjVuViewer._static_mounted:
            app.mount(
                "/static/djvu",
                StaticFiles(directory=self.config.images_path),
                name="djvu_images",
            )
            DjVuViewer._static_mounted = True

    def handle_exception(self, e: BaseException, trace: Optional[bool] = None):
        """Handles an exception by creating an error message.

        Args:
            e (BaseException): The exception to handle.
            trace (bool, optional): Whether to include the traceback in the error message. Default is False.
        """
        if trace:
            self.error_msg = str(e) + "\n" + traceback.format_exc()
        else:
            self.error_msg = str(e)
        logging.error(self.error_msg)

    def sanitize_path(self, path: str) -> str:
        """
        fix mediawiki path quirks e.g. with blanks
        """
        path = path.replace(" ", "_")
        return path

    def get_file_content(self, file: str) -> Tuple[str, bytes]:
        """
        Retrieves a content file (PNG, JPG, YAML, etc.) from the tarball

        Args:
            file (str): The full path in the format <DjVu name>/<file name>.
        """
        djvu_name, filename = file.split("/", 1)
        tarball_path = Path(self.config.tarball_path) / f"{djvu_name}.tar"
        file_content = Tarball.read_from_tar(tarball_path, filename)
        return filename, file_content

    def create_content_response(self, filename: str, file_content: bytes) -> Response:
        # Detect MIME type based on file extension
        media_type, _ = mimetypes.guess_type(filename)
        if media_type is None:
            media_type = "application/octet-stream"  # Default for unknown types

        content_response = Response(content=file_content, media_type=media_type)

        return content_response

    def get_tarball_response(self, path: str) -> FileResponse:
        """Serves the complete tarball for download."""
        path = self.sanitize_path(path)
        tarball_file = Path(self.config.tarball_path) / f"{Path(path).stem}.tar"

        if not tarball_file.exists():
            raise HTTPException(status_code=404, detail="Tarball not found")

        response = FileResponse(
            path=tarball_file,
            media_type="application/x-tar",
            filename=f"{Path(path).stem}.tar",  # Sets Content-Disposition automatically
        )
        return response

    def get_content(self, file: str) -> Response:
        """
        Retrieves a content file (PNG, JPG, YAML, etc.) from the tarball and serves it as a response.

        Args:
            file (str): The full path in the format <DjVu name>/<file name>.

        Returns:
            Response: The requested content file with the correct media type.
        """
        try:
            filename, file_content = self.get_file_content(file)
        except ValueError:
            raise HTTPException(
                status_code=400,
                detail="Invalid file path format. Expected <DjVu name>/<file name>.",
            )
        except FileNotFoundError:
            raise HTTPException(status_code=404, detail="Tarball not found")
        except KeyError:
            djvu_name, filename = file.split("/", 1)
            raise HTTPException(
                status_code=404,
                detail=f"File {filename} of DjVu {djvu_name} not found in tarball",
            )
        content_response = self.create_content_response(filename, file_content)
        return content_response

    def get_djvu_view_page(self, path: str, page_index: int) -> DjVuViewPage:
        """
        Helper function to fetch DjVu page data.

        Args:
            path (str): Path to the DjVu file (without page notation).
            page_index (int): Page number to display (1-based).

        Returns:
            DjVuViewPage: dataclass instance with file,page and image_url
        """
        path = self.sanitize_path(path)
        tarball_file = Path(self.config.tarball_path) / f"{Path(path).stem}.tar"
        yaml_file = f"{Path(path).stem}.yaml"

        if not tarball_file.exists():
            raise HTTPException(status_code=404, detail="Tarball not found")

        try:
            yaml_data = Tarball.read_from_tar(tarball_file, yaml_file).decode("utf-8")
            djvu_file = DjVuFile.from_yaml(yaml_data)  # @UndefinedVariable
        except Exception as ex:
            self.handle_exception(ex)
            raise HTTPException(
                status_code=500, detail=f"Error reading {yaml_file} from tarball"
            )

        page_count = len(djvu_file.pages)
        if page_index < 1 or page_index > page_count:
            raise HTTPException(status_code=404, detail=f"Page {page_index} not found")

        djvu_page = djvu_file.pages[page_index - 1]
        djvu_view_page = DjVuViewPage(file=djvu_file, page=djvu_page, base_path=path)
        return djvu_view_page

    def get_page4path(
        self, path: str, pageno: int, ext: str, scale: float = 1.0, quality: int = 85
    ) -> Response:
        """
        Fetches and displays a specific page of a DjVu file in the desired format.

        Args:
            path (str): The path to the DjVu document.
            pageno (int): The page number within the DjVu document.
            ext (str): The desired file extension for the page (e.g., "png", "jpg").
            scale (float, optional): The scale factor to apply to the image (0.0-1.0). Defaults to 1.0.
            quality (int, optional): The JPEG quality (1-100). Defaults to 85.

        Returns:
            Response: Response with the page content in the requested format.

        Raises:
            HTTPException: With status code 501 if the specified file extension is unsupported.
            HTTPException: With status code 500 if an error occurs while retrieving the page content.
        """
        # Check if the file extension is supported
        exts = ["png", "jpg"]
        if ext not in exts:
            msg = (
                f"Unsupported file extension: {ext}. Must be one of {', '.join(exts)}."
            )
            raise HTTPException(status_code=501, detail=msg)

        try:
            # Get the DjVu view page
            djvu_view_page = self.get_djvu_view_page(path, pageno)
            content_path = djvu_view_page.content_path

            # Get the original file content (PNG format)
            filename, file_content = self.get_file_content(content_path)

            # If JPG is requested, convert the PNG data to JPG
            if ext == "jpg":
                # Extract the DPI from the page metadata (assuming it's available)
                # If not available, you might need to add a parameter or use a default value
                dpi = getattr(
                    djvu_view_page.page, "dpi", 300
                )  # Default to 300 if not specified

                # Use ImageConverter to convert PNG to JPG
                converter = ImageConverter(file_content, dpi)
                file_content = converter.convert_to_jpg(scale=scale, quality=quality)

                # Update filename to reflect the JPG extension
                filename = filename.replace(".png", ".jpg")

            # Create and return the response with the appropriate content
            file_response = self.create_content_response(filename, file_content)
            return file_response
        except Exception as e:
            raise HTTPException(
                status_code=500, detail=f"Error retrieving page content: {str(e)}"
            )

    def create_page_dropdown(self, path, current_page, total_pages):
        """
        Create an HTML select dropdown for page navigation

        Args:
            path: Path of the DjVu file
            current_page: Currently displayed page number
            total_pages: Total number of pages in the document

        Returns:
            HTML select element with page options
        """
        options = []
        for page_num in range(1, total_pages + 1):
            selected = " selected" if page_num == current_page else ""
            options.append(f'<option value="{page_num}"{selected}>{page_num}</option>')

        options_html = "\n".join(options)

        select_html = f"""<select onchange="window.location.href='{self.url_prefix}/djvu/{path}?page='+this.value">
        {options_html}
    </select>"""

        return select_html

    def get_page(
        self, path: str, page_index: int, backlink: str = None
    ) -> HTMLResponse:
        """
        Fetches and renders an HTML page displaying the PNG image of the given DjVu file page from a tarball.
        """
        djvu_view_page = self.get_djvu_view_page(path, page_index)
        djvu_file = djvu_view_page.file
        image_url = djvu_view_page.image_url
        html_response = HTMLResponse(
            content=self.get_markup(
                path, page_index, len(djvu_file.pages), image_url, backlink
            )
        )
        return html_response

    def get_markup(
        self,
        path: str,
        page_index: int,
        total_pages: int,
        image_url: str,
        backlink: str = None,
    ) -> str:
        """
        Returns the HTML markup for displaying the DjVu page with navigation.

        Args:
            path (str): DjVu file path.
            page_index (int): Current page index.
            total_pages (int): Total number of pages in the DjVu document.
            image_url (str): URL to the PNG file.
            backlink (str, optional): URL to navigate back to. Defaults to None.

        Returns:
            str: HTML markup.
        """
        first_page = 1  # Fix: Pages start from 1
        last_page = total_pages  # Fix: Last page is total_pages, not total_pages - 1
        prev_page = max(first_page, page_index - 1)
        next_page = min(last_page, page_index + 1)
        fast_backward = max(first_page, page_index - 10)
        fast_forward = min(last_page, page_index + 10)
        select_markup = self.create_page_dropdown(path, page_index, total_pages)
        download_page_url = f"{self.url_prefix}{image_url}"
        download_tar_url = f"{self.url_prefix}/djvu/download/{path}"
        help_url = "https://github.com/WolfgangFahl/djvu-viewer/wiki/Help"
        # Add back button if backlink is provided
        back_button = ""
        if backlink:
            back_button = f'<a href="{backlink}" title="Back">⬅️</a> | '

        markup = f"""
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <title>DjVu Viewer</title>
            <style>
                body {{ font-family: Arial, sans-serif; text-align: center; }}
                img {{ max-width: 100%; height: auto; }}
                .nav {{ margin-top: 20px; }}
                .nav a {{ margin: 0 10px; text-decoration: none; font-weight: bold; font-size: 24px; }}
            </style>
        </head>
        <body>
            <div class="nav">
                {back_button}
                <a href="{self.url_prefix}/djvu/{path}?page={first_page}" title="First Page (1/{total_pages})">⏮</a>
                <a href="{self.url_prefix}/djvu/{path}?page={fast_backward}" title="Fast Backward (Jump -10 Pages)">⏪</a>
                <a href="{self.url_prefix}/djvu/{path}?page={prev_page}" title="Previous Page">⏴</a>
                <span>{select_markup} / {total_pages}</span>
                <a href="{self.url_prefix}/djvu/{path}?page={next_page}" title="Next Page">⏵</a>
                <a href="{self.url_prefix}/djvu/{path}?page={fast_forward}" title="Fast Forward (Jump +10 Pages)">⏩</a>
                <a href="{self.url_prefix}/djvu/{path}?page={last_page}" title="Last Page ({total_pages}/{total_pages})">⏭</a>
                | <a href="{download_page_url}" title="Page" download>⬇️</a>
                <a href="{download_tar_url}"  title="Tarball" download>📦</a>
                <a href="{help_url}" title="Help" target="_blank">❓</a>
            </div>
            <img src="{self.url_prefix}{image_url}" alt="DjVu Page {page_index}">
        </body>
        </html>
        """
        return markup



"""
Created on 2026-01-02

@author: wf
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Union

import requests
from ngwidgets.progress import Progressbar


@dataclass
class MediaWikiImage:
    """
    Represents a single image resource from MediaWiki.
    """

    name: str
    url: str
    mime: str
    size: int
    user: Optional[str] = None
    timestamp: Optional[datetime] = None
    description_url: Optional[str] = None
    height: Optional[int] = None
    width: Optional[int] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "MediaWikiImage":
        """
        Factory method to create an instance from a MediaWiki API result dict.
        """
        # Parse timestamp if available, usually ISO 8601 format like "2023-01-01T12:00:00Z"
        ts_str = data.get("timestamp")
        ts = None
        if ts_str:
            try:
                # Basic ISO parsing
                ts = datetime.strptime(ts_str, "%Y-%m-%dT%H:%M:%SZ")
            except ValueError:
                pass

        mw_image = cls(
            name=data.get("name") or data.get("title", "?"),
            url=data.get("url", ""),
            mime=data.get("mime", "application/octet-stream"),
            size=data.get("size", 0),
            user=data.get("user", None),
            timestamp=ts,
            description_url=data.get("descriptionurl"),
            height=data.get("height"),
            width=data.get("width"),
        )
        return mw_image


class MediaWikiImages:
    """
    Fetch images from a MediaWiki API via the 'allimages' list or by title.

    Example:
        client = MediaWikiImages(
            api_url="https://genwiki.genealogy.net/w/api.php",
            mime_types=("image/vnd.djvu", "image/x-djvu"),
            aiprop=("url", "mime", "size", "timestamp", "user"),
            timeout=20,
        )
        # Fetch list of images
        images = client.fetch_allimages(limit=120)

        # Fetch single image details
        single_img = client.fetch_image("File:Example.djvu")
    """

    def __init__(
        self,
        api_url: str,
        mime_types: Optional[Iterable[str]] = None,
        aiprop: Optional[Iterable[str]] = None,
        timeout: int = 30,
        session: Optional[requests.Session] = None,
    ):
        """
        Args:
            api_url: Full API endpoint, e.g. 'https://genwiki.genealogy.net/w/api.php'
            mime_types: MIME types to filter by (joined with '|'). e.g. ("image/jpeg",).
            aiprop: Properties to request. Defaults to url, mime, size, timestamp, user, dimensions.
            timeout: Per-request timeout in seconds.
            session: Optional requests.Session to reuse connections.
        """
        self.api_url = api_url
        self.timeout = timeout
        self.session = session or requests.Session()

        # Default filters
        self.mime_types = tuple(mime_types) if mime_types else ()

        # Default properties if none provided
        if aiprop is None:
            self.aiprop = ("url", "mime", "size", "timestamp", "user", "dimensions")
        else:
            self.aiprop = tuple(aiprop)

    def fetch_image(self, title: str) -> Optional[MediaWikiImage]:
        """
        Retrieve a single image by its file title (e.g., 'File:Example.jpg').

        Args:
            title: The exact page title of the file.

        Returns:
            MediaWikiImage object if found, None otherwise.
        """
        if not ":" in title:
            title = "File:" + title
        params = {
            "action": "query",
            "prop": "imageinfo",
            "titles": title,
            "format": "json",
            "iiprop": "|".join(self.aiprop),
        }

        data = self._make_request(params)
        pages = data.get("query", {}).get("pages", {})
        mw_image = None
        for page_id, page_data in pages.items():
            # If page_id is negative (e.g., "-1"), the page does not exist
            if int(page_id) < 0:
                continue

            imageinfo = page_data.get("imageinfo", [])
            if imageinfo:
                # Merge page title into the image info dict to match 'allimages' structure
                info_dict = imageinfo[0]
                info_dict["title"] = page_data.get("title")
                mw_image = MediaWikiImage.from_dict(info_dict)

        return mw_image

    def fetch_allimages(
        self,
        limit: int,
        per_request: int = 50,
        extra_params: Optional[Dict[str, str]] = None,
        as_objects: bool = False,
        progressbar: Optional[Progressbar] = None,
    ) -> Union[List[MediaWikiImage], List[Dict]]:
        """
        Retrieve up to 'limit' images.

        Args:
            limit: Maximum number of image records to return.
            per_request: Page size for each API call (max 50 normally, 500 for bots).
            extra_params: Extra query params to merge into the request.
            as_objects: If True, returns List[MediaWikiImage]. If False, returns List[Dict].

        Returns:
            List of MediaWikiImage objects or dictionaries.
        """
        if progressbar:
            progressbar.total = limit or 0  # Set total if known

        results = []
        remaining = max(0, int(limit))
        if remaining == 0:
            return results

        base_params = {
            "action": "query",
            "list": "allimages",
            "format": "json",
        }

        # Apply Configuration filters
        if self.mime_types:
            base_params["aimime"] = "|".join(self.mime_types)
        if self.aiprop:
            base_params["aiprop"] = "|".join(self.aiprop)

        if extra_params:
            base_params.update(extra_params)

        continue_params: Dict[str, str] = {}

        while remaining > 0:
            params = dict(base_params)
            params.update(continue_params)
            params["ailimit"] = min(per_request, remaining)

            data = self._make_request(params)

            # Extract list
            images_raw = data.get("query", {}).get("allimages", [])
            if not images_raw:
                break

            # Append up to 'remaining'
            take = min(remaining, len(images_raw))
            batch = images_raw[:take]

            if as_objects:
                results.extend([MediaWikiImage.from_dict(img) for img in batch])
            else:
                results.extend(batch)
            # Update progress bar
            if progressbar:
                progressbar.update(len(batch))

            remaining -= take

            # Handle pagination
            cont = data.get("continue")
            if remaining > 0 and cont:
                continue_params = cont
            else:
                break

        return results

    def _make_request(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """
        Helper to execute the request and handle basic errors.
        """
        resp = self.session.get(
            self.api_url, params=params, timeout=self.timeout, allow_redirects=True
        )
        resp.raise_for_status()
        data = resp.json()

        if "error" in data:
            code = data["error"].get("code", "unknown")
            info = data["error"].get("info", "")
            raise RuntimeError(f"MediaWiki API error: {code} - {info}")

        return data



"""
Created on 2024-08-15

@author: wf
Refactored to Focus on DjVu functionality
"""

from argparse import Namespace
from typing import Dict

from djvuviewer.djvu_catalog import DjVuCatalog
from djvuviewer.djvu_config import DjVuConfig
from djvuviewer.djvu_debug import DjVuDebug
from djvuviewer.djvu_viewer import DjVuViewer
from djvuviewer.djvu_wikimages import DjVuMediaWikiImages
from djvuviewer.version import Version
from ngwidgets.input_webserver import InputWebserver, InputWebSolution
from ngwidgets.login import Login
from ngwidgets.sso_users_solution import SsoSolution
from ngwidgets.webserver import WebserverConfig
from ngwidgets.widgets import Link
from nicegui import Client, app, ui
from starlette.responses import FileResponse, HTMLResponse
from wikibot3rd.sso_users import Sso_Users

from djvuviewer.djvu_context import DjVuContext


class DjVuViewerWebServer(InputWebserver):
    """WebServer class that manages the server and handles DjVu operations."""

    @classmethod
    def get_config(cls) -> WebserverConfig:
        copy_right = "(c)2024-2026 Wolfgang Fahl"
        config = WebserverConfig(
            copy_right=copy_right,
            # Ideally, the Version class should also belong to a package named 'djvu_viewer' or similar
            version=Version(),
            default_port=9840,
            short_name="djvuviewer",
        )
        server_config = WebserverConfig.get(config)
        server_config.solution_class = DjVuSolution
        return server_config

    def __init__(self):
        """Constructs all the necessary attributes for the WebServer object."""
        InputWebserver.__init__(self, config=DjVuViewerWebServer.get_config())
        self.users = Sso_Users(self.config.short_name)
        self.login = Login(self, self.users)
        self.djvu_config = DjVuConfig.get_instance()


        @ui.page("/")
        async def home(client: Client):
            # Default to catalog as it is the primary remaining feature
            return await self.page(client, DjVuSolution.djvu_catalog)

        @ui.page("/djvu/catalog")
        async def djvu_catalog(client: Client):
            return await self.page(client, DjVuSolution.djvu_catalog)

        @ui.page("/djvu/browse")
        async def djvu_browse(client: Client):
            return await self.page(client, DjVuSolution.djvu_browse)

        @ui.page("/djvu/debug/{page_title:str}")
        async def djvu_debug_route(client: Client, page_title: str) -> HTMLResponse:
            """Route for DjVu debug page"""
            return await self.page(client, DjVuSolution.djvu_debug, page_title)

        @ui.page("/login")
        async def login(client: Client) -> None:
            return await self.page(client, DjVuSolution.show_login)

        # Add the static files route for serving the backup files
        app.add_static_files("/backups", self.djvu_config.backup_path)

        @app.get("/djvu/content/{file:path}")
        def get_content(file: str) -> FileResponse:
            """
            Serves content from a wrapped DjVu file.

            Args:
                file (str): The full path  <DjVu name>/<file name>.

            Returns:
                FileResponse: The requested content file (PNG, JPG, YAML, etc.).
            """
            file_response = self.djvu_viewer.get_content(file)
            return file_response

        @app.get("/djvu/download/{path:path}")
        def download_tarball(path: str) -> FileResponse:
            """
            Serves the complete tarball for download.

            Args:
                path (str): The path to the DjVu document.
            """
            response = self.djvu_viewer.get_tarball_response(path)
            return response

        @app.get("/djvu/{path:path}/page/{scale:float}/{pageno:int}.{ext:str}")
        def get_djvu_page_with_scale(
            path: str,
            pageno: int,
            scale: float = 1.0,
            ext: str = "png",
            quality: int = 85,
        ) -> FileResponse:
            """
            Fetches and displays a specific PNG page of a DjVu file.

            Args:
                path (str): The path to the DjVu document.
                pageno (int): The page number within the DjVu document.
                scale(float,optional): the scale of the jpg impage
                ext (str): The desired file extension for the page ("png" or "jpg").
                quality (int, optional): The desired jpg quality - default:85
            """
            file_response = self.djvu_viewer.get_page4path(
                path, pageno, ext=ext, scale=scale, quality=quality
            )
            return file_response

        @app.get("/djvu/{path:path}/page/{pageno:int}.{ext:str}")
        def get_djvu_page(
            path: str,
            pageno: int,
            scale: float = 1.0,
            ext: str = "png",
            quality: int = 85,
        ) -> FileResponse:
            """
            Fetches and displays a specific PNG page of a DjVu file.

            Args:
                path (str): The path to the DjVu document.
                pageno (int): The page number within the DjVu document.
                scale(float,optional): the scale of the jpg impage
                ext (str): The desired file extension for the page ("png" or "jpg").
                quality (int, optional): The desired jpg quality - default:85
            """
            file_response = self.djvu_viewer.get_page4path(
                path, pageno, ext=ext, scale=scale, quality=quality
            )
            return file_response

        @app.get("/djvu/{path:path}")
        def display_djvu(
            path: str, page: int = 1, backlink: str = None
        ) -> HTMLResponse:
            """
            Fetches and displays a specific PNG page of a DjVu file.

            Args:
                path: Path to the DjVu file
                page: Page number to display (default: 1)
                backlink: Optional URL to return to (default: None)
            """
            html_response = self.djvu_viewer.get_page(
                path=path, page_index=page, backlink=backlink
            )
            return html_response

    def authenticated(self) -> bool:
        """
        check authentication
        """
        allow = self.login.authenticated()
        return allow

    def configure_run(self):
        """
        configure me
        """
        super().configure_run()
        djvu_cmd_args = Namespace(
            # From BaseCmd
            debug=self.args.debug,
            verbose=self.args.verbose,
            force=self.args.force,
            quiet=self.args.quiet,
            # From DjVu config
            images_path=self.djvu_config.images_path,
            db_path=self.djvu_config.db_path,
            backup_path=self.djvu_config.backup_path,
            container_name=self.djvu_config.container_name,
            # DjVu defaults
            batch_size=100,
            command="bundle",
            limit=10000000,
            limit_gb=16,
            max_errors=1.0,
            sleep=2.0,
            max_workers=None,
            output_path=None,
            pngmode="pil",
            serial=False,
            sort="asc",
            url=None,
            cleanup=False,
            script=False,
            dry_run=False,
        )
        self.context=DjVuContext(self.djvu_config,djvu_cmd_args)
        # make helper classes available
        self.djvu_viewer = DjVuViewer(app=app, config=self.djvu_config)
        # Initialize MediaWiki clients if using API mode
        self.mw_client_base = DjVuMediaWikiImages.get_mediawiki_images_client(
            self.djvu_config.base_url
        )
        self.mw_client_new = DjVuMediaWikiImages.get_mediawiki_images_client(
            self.djvu_config.new_url
        )

class DjVuSolution(InputWebSolution):
    """
    the DjVuViewer solution
    """

    def __init__(self, webserver: DjVuViewerWebServer, client: Client):
        """
        Initialize the solution

        Args:
            webserver (DjVuViewerWebServer): The webserver instance associated with this solution.
            client (Client): The client instance this context is associated with.
        """
        super().__init__(webserver, client)
        self.djvu_config = webserver.djvu_config

    def add_links(self, view_record: Dict[str, any], filename: str):
        """
        Add the DjVu links.
        """
        config = self.djvu_config
        if filename:
            wiki_url = self.djvu_config.wiki_fileurl(filename)
            view_record["wiki"] = Link.create(url=wiki_url, text=filename)

            if config.new_url:
                new_url = self.djvu_config.wiki_fileurl(filename, new=True)
                view_record["new"] = Link.create(url=new_url, text=filename)
                backlink = self.djvu_config.wiki_fileurl(
                    filename, new=True, quoted=True
                )
            backparam = f"?backlink={backlink}" if backlink else ""
            local_url = f"{config.url_prefix}/djvu/{filename}{backparam}"
            view_record["tarball"] = Link.create(url=local_url, text=filename)

            debug_url = f"{config.url_prefix}/djvu/debug/{filename}"
            view_record["debug"] = Link.create(url=debug_url, text="🔍")

    def configure_menu(self):
        """
        setup the menu
        """
        InputWebSolution.configure_menu(self)
        self.login = self.webserver.login
        self.sso_solution = SsoSolution(webserver=self.webserver)
        self.sso_solution.configure_menu()
        # icons from https://fonts.google.com/icons
        # if self.webserver.authenticated():
        #    self.link_button(name="wikis", icon_name="menu_book", target="/wikis")

        with self.header:
            self.link_button("DjVu Tarballs", "/djvu/catalog", "library_books")
            self.link_button("DjVu Wiki Images", "/djvu/browse", "image")

    async def show_login(self):
        """Show login page"""
        await self.login.login(self)

    async def djvu_debug(self, page_title: str):
        """Show the DjVu Debug page"""

        def show():
            debug_view = DjVuDebug(
                self,
                context=self.webserver.context,
                page_title=page_title,
            )
            debug_view.setup_ui()

        await self.setup_content_div(show)

    async def djvu_modal_catalog(self, browse_wiki: bool = False):
        """Show the DjVu Catalog page"""

        def show():
            self.djvu_catalog_view = DjVuCatalog(
                self,
                config=self.webserver.djvu_config,
                browse_wiki=browse_wiki,
            )
            self.djvu_catalog_view.setup_ui()

        await self.setup_content_div(show)

    async def djvu_catalog(self):
        await self.djvu_modal_catalog(browse_wiki=False)

    async def djvu_browse(self):
        await self.djvu_modal_catalog(browse_wiki=True)



"""
Created on 04.01.2026

@author: wf
"""
from argparse import Namespace

from basemkit.profiler import Profiler
from djvuviewer.djvu_actions import DjVuActions
from djvuviewer.djvu_config import DjVuConfig
from djvuviewer.djvu_manager import DjVuManager
from djvuviewer.djvu_processor import DjVuProcessor


class DjVuContext:
    """
    a Context for working with DjVu files and actions
    """
    def __init__(self,config:DjVuConfig,args:Namespace):
        self.config=config
        self.args=args
        # Initialize manager and processor
        self.dvm = DjVuManager(config=self.config)
        self.dproc = DjVuProcessor(
            debug=self.args.debug,
            verbose=self.args.verbose,
            batch_size=self.args.batch_size,
            limit_gb=self.args.limit_gb,
            max_workers=self.args.max_workers,
            pngmode=self.args.pngmode,
        )

        # Initialize actions handler
        self.actions = DjVuActions(
            config=self.config,
            args=self.args,
            dvm=self.dvm,
            dproc=self.dproc,
            images_path=self.args.images_path,
            output_path=self.args.output_path,
            debug=self.args.debug,
            verbose=self.args.verbose,
            force=self.args.force,
        )

        self.profiler = Profiler(self.args.command)



"""
Created on 2024-09-12

@author: wf
"""

from lodstorage.query import Query, QueryManager


class MultiLanguageQueryManager:
    """
    Query manager for multiple languages
    """

    def __init__(
        self,
        yaml_path: str,
        languages: list = ["sql", "sparql", "ask"],
        debug: bool = False,
    ):
        self.languages = languages
        self.debug = debug
        self.qms = {}
        for lang in languages:
            qm = QueryManager(
                lang=lang, queriesPath=yaml_path, with_default=False, debug=self.debug
            )
            self.qms[lang] = qm

        self.query_names = []
        for qm in self.qms.values():
            self.query_names.extend(list(qm.queriesByName.keys()))

    def query4Name(self, name: str) -> Query:
        """ """
        result = None
        for qm in self.qms.values():
            if name in qm.queriesByName:
                result = qm.queriesByName[name]
                break
        return result



"""
Created on 2025-02-25

@author: wf
"""

from argparse import ArgumentParser, Namespace
import argparse
import logging
from typing import List, Optional

from basemkit.base_cmd import BaseCmd
from djvuviewer.djvu_config import DjVuConfig
from djvuviewer.version import Version

from djvuviewer.djvu_context import DjVuContext


class DjVuCmd(BaseCmd):
    """
    Command-line interface for DjVu file processing and conversion.

    Provides commands for:
    - catalog: Scan and catalog DjVu files
    - convert: Convert DjVu files to PNG format
    - dbupdate: Update database with processed file metadata
    - initdb: Initialize the database schema
    """

    def __init__(self, args: argparse.Namespace):
        """
        Initialize the DjVu command-line interface.

        Args:
            args: Parsed command-line arguments
        """
        super().__init__(Version())
        self.args = args
        self.config = DjVuConfig.get_instance()
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        )

    def add_arguments(self, parser: ArgumentParser) -> ArgumentParser:
        """
        Add command-specific arguments to the argument parser.

        Args:
            parser: ArgumentParser to add arguments to

        Returns:
            The modified ArgumentParser
        """
        super().add_arguments(parser)

        parser.add_argument(
            "--images-path",
            default=self.config.images_path,
            help="Base path for DjVu files e.g. images directory of a wiki",
        )
        parser.add_argument(
            "--batch-size",
            type=int,
            default=100,
            help="Number of pages to process in each batch (default: %(default)s)",
        )
        parser.add_argument(
            "--command",
            choices=["bundle", "catalog", "convert", "dbupdate", "initdb"],
            required=True,
            help="Command to execute",
        )
        parser.add_argument(
            "--db-path",
            default=self.config.db_path,
            help="Path to the database (default: %(default)s)",
        )
        parser.add_argument(
            "--limit",
            type=int,
            default=10000000,
            help="Maximum number of pages to process (default: %(default)s)",
        )
        parser.add_argument(
            "--limit_gb",
            type=int,
            default=16,
            help="Memory limit in GB (default: %(default)s)",
        )
        parser.add_argument(
            "--max-errors",
            type=float,
            default=1.0,
            help="Maximum allowed error percentage before skipping database update (default: %(default)s)",
        )
        parser.add_argument(
            "--sleep",
            type=float,
            default=2.0,
            help="number of seconds to sleep before trying to move to avoid actimeo issues on CIFS mounts (default: %(default)s)",
        )
        parser.add_argument(
            "--max-workers",
            type=int,
            default=None,
            help="Maximum number of worker threads (default: CPU count * 4)",
        )
        parser.add_argument(
            "--output-path",
            help="Path for PNG files",
        )
        parser.add_argument(
            "--pngmode",
            choices=["cli", "pil"],
            default="pil",
            help="PNG generation mode: cli (ddjvu command) or pil (Python Imaging Library) (default: %(default)s)",
        )
        parser.add_argument(
            "--serial",
            action="store_true",
            help="Use serial processing instead of parallel",
        )
        parser.add_argument(
            "--sort",
            choices=["asc", "desc"],
            default="asc",
            help="Sort by page count (asc=smallest first, default: %(default)s)",
        )
        parser.add_argument(
            "--url",
            help="Process a single DjVu file (only valid in convert mode)",
        )
        # Bundle-specific arguments
        parser.add_argument(
            "--backup-path",
            default=self.config.backup_path,
            help="Path for backup ZIP files default: %(default)s)",
        )
        parser.add_argument(
            "--cleanup",
            action="store_true",
            help="Remove thumbnails during bundling",
        )
        parser.add_argument(
            "--script",
            action="store_true",
            help="Generate bash script instead of executing",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Show operations without executing (implies --script)",
        )
        parser.add_argument(
            "--container-name",
            default=self.config.container_name,
            help="MediaWiki container for maintenance call default: %(default)s)",
        )

        return parser

    def handle_args(self, args: Namespace) -> bool:
        """
        Handle the command-line arguments and execute the requested command.

        Sets up the configuration and dispatches to the appropriate command handler.

        Args:
            args: Parsed command-line arguments

        Returns:
            True if handled (no further processing needed)
        """
        handled = super().handle_args(args)
        # Configure paths
        self.config.db_path = self.args.db_path
        self.config.images_path = self.args.images_path
        self.config.backup_path = self.args.backup_path
        self.config.container_name = self.args.container_name
        self.context=DjVuContext(self.config,self.args)
        self.actions=self.context.actions
        self.profiler=self.context.profiler

        # Dispatch to command handler
        command_handlers = {
            "bundle": self.bundle,
            "catalog": self.catalog,
            "convert": self.convert,
            "dbupdate": self.dbupdate,
            "initdb": self.initdb,
        }

        handler = command_handlers.get(self.args.command)
        if handler:
            handler()
            handled = True
        else:
            print(f"unknown command {self.args.command}")
        return handled

    def bundle(self):
        """
        Execute the bundle command.

        Converts indirect/multi-file DjVu files to bundled format.
        """
        self.actions.bundle_djvu_files()
        self.actions.report_errors(profiler_time_func=self.profiler.time)

    def catalog(self) -> None:
        """
        Execute the catalog command.

        Scans DjVu files and stores their metadata in the database.
        """
        self.actions.catalog_and_store(limit=self.args.limit, sample_record_count=1)
        self.actions.report_errors(profiler_time_func=self.profiler.time)

    def convert(self) -> None:
        """
        Execute the convert command.

        Converts DjVu files to PNG format using database records.
        """
        self.actions.convert_from_database(serial=self.args.serial, url=self.args.url)
        self.actions.report_errors(profiler_time_func=self.profiler.time)

    def dbupdate(self) -> None:
        """
        Execute the dbupdate command.

        Updates the DjVu database with metadata from processed files.
        """
        self.actions.update_from_database(
            max_errors=self.args.max_errors, url=self.args.url
        )
        self.actions.report_errors(profiler_time_func=self.profiler.time)

    def initdb(self) -> None:
        """
        Execute the initdb command.

        Initializes the database schema with sample records.
        """
        self.actions.init_database()
        self.actions.report_errors(profiler_time_func=self.profiler.time)


def main(argv: Optional[List[str]] = None) -> int:
    """
    Main entry point for the DjVu command-line tool.

    Args:
        argv: Command-line arguments (defaults to sys.argv if None)

    Returns:
        Exit code (0 for success, non-zero for failure)
    """
    exit_code = DjVuCmd.main(argv)
    return exit_code


if __name__ == "__main__":
    main()



"""
Created on 2026-01-03

@author: wf
"""

import io
import os
import re
import shlex
import shutil
from datetime import datetime
import subprocess
import tarfile
import tempfile
import time
import zipfile
from pathlib import Path
from typing import List, Optional

from basemkit.shell import Shell
from PIL import Image

from djvuviewer.djvu_config import DjVuConfig
from djvuviewer.djvu_core import DjVuFile
from djvuviewer.tarball import Tarball


class DjVuBundle:
    """
    DjVu bundle handling with validation and error collection.
    """

    def __init__(
        self, djvu_file: DjVuFile, config: DjVuConfig = None, debug: bool = False
    ):
        """
        Initialize DjVuBundle with a DjVuFile instance.

        Args:
            djvu_file: The DjVuFile metadata
            config: configuration
            debug: if True
        """
        self.djvu_file = djvu_file
        if config is None:
            config = DjVuConfig.get_instance()
        self.config = config
        self.debug = debug
        self.errors: List[str] = []
        self.shell = Shell()
        self.djvu_dump_log = None

    @classmethod
    def from_tarball(cls, tar_file: str, with_check: bool = True) -> "DjVuBundle":
        """
        Create a DjVuBundle from a tarball file.

        Args:
            tar_file: Path to the tar archive
            with_check: If True, automatically run validation checks

        Returns:
            DjVuBundle: Instance with loaded metadata and optional validation

        Example:
            bundle = DjVuBundle.from_tarball("document.tar")
            if not bundle.is_valid():
                print(bundle.get_error_summary())
        """
        tarball_path = Path(tar_file)

        # Find YAML file in tarball
        with tarfile.open(tar_file) as tar:
            yaml_files = [m.name for m in tar.getmembers() if m.name.endswith(".yaml")]

        if not yaml_files:
            # Create a minimal bundle with an error
            bundle = cls(DjVuFile(pages=[]))
            bundle.errors.append(f"No YAML metadata file found in tarball: {tar_file}")
            return bundle

        if len(yaml_files) > 1:
            # Use first one but warn
            bundle = cls(DjVuFile.from_tarball(tarball_path, yaml_files[0]))
            bundle.errors.append(
                f"Found {len(yaml_files)} YAML files, expected 1. Using {yaml_files[0]}"
            )
        else:
            bundle = cls(DjVuFile.from_tarball(tarball_path, yaml_files[0]))

        if with_check:
            bundle.check_tarball(tar_file)

        return bundle

    def _add_error(self, message: str):
        """Add an error message to the error list."""
        self.errors.append(message)

    def check_tarball(self, tar_file: str, relurl: Optional[str] = None):
        """
        Verify that a tar file exists and contains expected contents with correct dimensions.
        Collects errors instead of raising exceptions.

        Args:
            tar_file: Path to the tar file to validate
            relurl: Optional relative URL for error context
        """
        context = f" for {relurl}" if relurl else ""

        # Check file exists
        if not os.path.isfile(tar_file):
            self._add_error(f"Expected tar file '{tar_file}' was not created{context}")
            return

        try:
            with tarfile.open(tar_file) as tar:
                members = tar.getmembers()

                # Check tar is not empty
                if len(members) == 0:
                    self._add_error(f"Tar file '{tar_file}' is empty{context}")
                    return

                # Find PNG files
                png_files = [m.name for m in members if m.name.endswith(".png")]
                if len(png_files) == 0:
                    self._add_error(f"No PNG files found in tar{context}")

                # Check for YAML file
                yaml_files = [m.name for m in members if m.name.endswith(".yaml")]
                if len(yaml_files) == 0:
                    self._add_error(f"No YAML file found in tar{context}")
                    return
                elif len(yaml_files) > 1:
                    self._add_error(
                        f"Expected 1 YAML file in tar, found {len(yaml_files)}{context}"
                    )

                # Create dimension mapping from metadata
                page_dimensions = {
                    i + 1: (page.width, page.height)
                    for i, page in enumerate(self.djvu_file.pages)
                }

                # Verify PNG dimensions match metadata
                tarball_path = Path(tar_file)
                for png_file in sorted(png_files):
                    basename = os.path.basename(png_file)
                    match = re.search(r"(\d+)\.png$", basename)

                    if not match:
                        self._add_error(
                            f"Could not extract page number from PNG: {png_file}{context}"
                        )
                        continue

                    page_num = int(match.group(1))
                    if page_num not in page_dimensions:
                        self._add_error(
                            f"PNG {png_file} references page {page_num} not in YAML{context}"
                        )
                        continue

                    expected_width, expected_height = page_dimensions[page_num]

                    try:
                        png_data = Tarball.read_from_tar(tarball_path, png_file)
                        with Image.open(io.BytesIO(png_data)) as img:
                            actual_width, actual_height = img.size

                        if actual_width != expected_width:
                            self._add_error(
                                f"PNG {png_file} width mismatch: expected {expected_width}, "
                                f"got {actual_width}{context}"
                            )

                        if actual_height != expected_height:
                            self._add_error(
                                f"PNG {png_file} height mismatch: expected {expected_height}, "
                                f"got {actual_height}{context}"
                            )
                    except Exception as e:
                        self._add_error(
                            f"Failed to read/validate PNG {png_file}: {e}{context}"
                        )

        except tarfile.TarError as e:
            self._add_error(f"Failed to open tar file '{tar_file}': {e}{context}")
        except Exception as e:
            self._add_error(
                f"Unexpected error checking tar file '{tar_file}': {e}{context}"
            )

    def get_part_filenames(self) -> List[str]:
        """
        get a list of my part file names
        """
        if not self.djvu_dump_log:
            self.djvu_dump()
        part_files = []
        for line in self.djvu_dump_log.split("\n"):
            match = re.search(r"^\s+(.+\.(?:djvu|djbz))\s+->", line)
            if match:
                part_files.append(match.group(1))
        return part_files

    def djvu_dump(self) -> str:
        """
        Run djvudump on self.djvu_file.djvu_path and return output.
        Adds error to self.errors on failure.

        Returns:
            djvudump output string (empty on error)
        """
        djvu_path = self.djvu_file.path

        if not os.path.exists(djvu_path):
            self._add_error(f"File not found: {djvu_path}")
            return ""

        cmd = f"djvudump {shlex.quote(djvu_path)}"
        result = self.run_cmd(cmd, "djvudump failed")

        if result.returncode == 0:
            self.djvu_dump_log = result.stdout
        return self.djvu_dump_log

    def finalize_bundling(self, zip_path: str, bundled_path: str, sleep: float = 1.0):
        """
        Finalize bundling by removing the original main file and
        all zipped component files then move the bundled_path file to the original.

        Args:
            zip_path: Path to the backup ZIP file (for verification)
            bundled_path: Path to the new bundled DjVu file
        """
        djvu_path = self.djvu_file.path

        # Verify backup ZIP exists before proceeding
        if not os.path.exists(zip_path):
            self._add_error(f"Backup ZIP not found: {zip_path}")
            return

        # Verify bundled file exists
        if not os.path.exists(bundled_path):
            self._add_error(f"Bundled file not found: {bundled_path}")
            return

        # Get directory of original file
        djvu_dir = os.path.dirname(djvu_path)

        # Get list of component files to remove
        part_files = self.get_part_filenames()

        try:
            original_stat = os.stat(djvu_path)
            original_atime = original_stat.st_atime
            original_mtime = original_stat.st_mtime
            if self.debug:
                print(f"Preserving timestamps - atime: {original_atime}, mtime: {original_mtime}")

            # Remove component parts
            for part_file in part_files:
                part_path = os.path.join(djvu_dir, part_file)
                if os.path.exists(part_path):
                    os.remove(part_path)
                    if self.debug:
                        print(f"Removed component: {part_path}")

            # Before attempting finalization
            if not os.path.exists(bundled_path):
                self._add_error(f"Bundled file missing: {bundled_path}")
                return

            if not os.access(djvu_dir, os.W_OK):
                self._add_error(
                    f"No write permission in directory: {djvu_dir}\n"
                    f"Try: sudo chmod g+w {djvu_dir}"
                )
                return

            if self.debug:
                print(f"trying to\nmv {bundled_path} {djvu_path}")
            if self.move_file(bundled_path, djvu_path):
                # Restore original timestamps
                os.sync()
                print(f"Sleeping {sleep} secs")
                time.sleep(sleep)
                os.utime(djvu_path, (original_atime, original_mtime))
                if self.debug:
                    print(f"Restored timestamps to {djvu_path}")

            # Update the DjVuFile object to reflect it's now bundled
            self.djvu_file.bundled = True

        except Exception as e:
            self._add_error(f"Error during finalization: {e}")

    @property
    def basename(self)->str:
        djvu_path = self.djvu_file.path

        # Get base filename without extension
        basename = os.path.basename(djvu_path)
        return basename

    @property
    def backup_file(self)->str:
        stem = os.path.splitext(self.basename)[0]

        # Create backup ZIP path
        backup_file = os.path.join(self.config.backup_path, f"{stem}.zip")
        return backup_file

    def create_backup_zip(self) -> str:
        """
        Create a ZIP backup of all unbundled DjVu files.

        Returns:
            Path to created backup ZIP file
        """
        if self.djvu_file.bundled:
            raise ValueError(f"File {self.djvu_file.path} is already bundled")

        djvu_path = self.djvu_file.path

        backup_file=self.backup_file

        # Get list of page files
        part_files = self.get_part_filenames()
        djvu_dir = os.path.dirname(djvu_path)

        # Create ZIP archive
        with zipfile.ZipFile(backup_file, "w", zipfile.ZIP_DEFLATED) as zipf:
            # Add main index file
            zipf.write(djvu_path, self.basename)

            # Add each page file
            for part_file in part_files:
                part_path = os.path.join(djvu_dir, part_file)
                if os.path.exists(part_path):
                    zipf.write(part_path, part_file)
                else:
                    self.errors.append(Exception(f"missing {part_path}"))

        return backup_file

    def run_cmd(self, cmd: str, error_msg: str = None) -> subprocess.CompletedProcess:
        """Run shell command with error handling."""
        result = self.shell.run(cmd, text=True, debug=self.debug)

        if result.returncode != 0:
            msg = error_msg or f"Command failed: {cmd}"
            if self.debug:
                print(f"{result.stdout}")
            self._add_error(f"{msg}\n{result.stderr}")

        return result

    def move_file(self, src: str, dst: str) -> bool:
        """Move file using copy+delete pattern for better reliability"""
        try:
            # First copy the file
            shutil.copy2(src, dst)  # copy2 preserves metadata
            if self.debug:
                print(f"Copied: {src} → {dst}")

            # Then remove the source
            os.remove(src)
            if self.debug:
                print(f"Removed source: {src}")

            return True

        except PermissionError as e:
            if self.debug:
                print(f"Permission error moving {src} → {dst}: {e}")
            self._add_error(f"Permission error: {e}")
            return False

        except FileNotFoundError as e:
            if self.debug:
                print(f"File not found {src} → {dst}: {e}")
            self._add_error(f"File not found: {src}")
            return False

        except Exception as e:
            if self.debug:
                print(f"Failed to move {src} → {dst}: {e}")
            self._add_error(f"Move failed: {e}")
        return False

    def generate_bundling_script(self) -> str:
        """
        Generate a complete bash script for the bundling process.

        Returns:
            The bash script as a string
        """
        djvu_path = self.djvu_file.path
        djvu_dir = os.path.dirname(djvu_path)
        basename = os.path.basename(djvu_path)
        stem = os.path.splitext(basename)[0]

        # Get part files
        part_files = self.get_part_filenames()

        # Define paths
        backup_file = os.path.join(self.config.backup_path, f"{stem}.zip")
        bundled_file = os.path.join(djvu_dir, f"{stem}_bundled.djvu")

        # Build script content
        script_lines = [
            "#!/bin/bash",
            "# DjVu bundling script",
            f"# Generated for: {djvu_path}",
            f"# Date: {datetime.now().isoformat()}",
            "",
            "set -e  # Exit on error",
            "",
            "# Define variables",
            f"DJVU_PATH={shlex.quote(djvu_path)}",
            f"DJVU_DIR={shlex.quote(djvu_dir)}",
            f"BASENAME={shlex.quote(basename)}",
            f"BACKUP_FILE={shlex.quote(backup_file)}",
            f"BUNDLED_FILE={shlex.quote(bundled_file)}",
            "",
            "# Step 1: Create backup ZIP",
            f'cd "$DJVU_DIR"',
            f"echo 'Creating backup ZIP...'",
            f'zip -j "$BACKUP_FILE" "$BASENAME" \\',
        ]

        # Add part files to zip command
        for i, part_file in enumerate(part_files):
            is_last = i == len(part_files) - 1
            line = f"  {shlex.quote(part_file)}"
            if not is_last:
                line += " \\"
            script_lines.append(line)

        script_lines.extend(
            [
                "",
                "# Step 2: Verify backup was created",
                'if [ ! -f "$BACKUP_FILE" ]; then',
                "  echo 'Error: Backup ZIP not created'",
                "  exit 1",
                "fi",
                "echo 'Backup created: '$BACKUP_FILE",
                "",
                "# Step 3: Convert to bundled format",
                "echo 'Converting to bundled format...'",
                'djvmcvt -b "$DJVU_PATH" "$BUNDLED_FILE"',
                "",
                "# Step 4: Verify bundled file was created",
                'if [ ! -f "$BUNDLED_FILE" ]; then',
                "  echo 'Error: Bundled file not created'",
                "  exit 1",
                "fi",
                "echo 'Bundled file created: '$BUNDLED_FILE",
                "",
                "# Step 5: Sleep for CIFS sync (if needed)",
                "sleep 1",
                "",
                "# Step 6: Remove original files",
                "echo 'Removing original files...'",
                f'rm -f "$DJVU_PATH"',
            ]
        )

        # Add removal of part files
        for part_file in part_files:
            script_lines.append(
                f"rm -f {shlex.quote(os.path.join(djvu_dir, part_file))}"
            )

        script_lines.extend(
            [
                "",
                "# Step 7: Move bundled file to original location",
                "echo 'Moving bundled file to original location...'",
                'mv "$BUNDLED_FILE" "$DJVU_PATH"',
                "",
                "echo 'Bundling complete!'",
                f"echo 'Backup saved at: '$BACKUP_FILE",
            ]
        )
        # MediaWiki maintenance call if container is configured
        if hasattr(self.config, "container_name") and self.config.container_name:
            filename = os.path.basename(djvu_path)
            docker_cmd = f"docker exec {self.config.container_name} php maintenance/refreshImageMetadata.php --force --mime=image/vnd.djvu --start={filename} --end={filename}"
            script_lines.extend([docker_cmd])

        return "\n".join(script_lines) + "\n"

    def convert_to_bundled(self, output_path: str = None) -> str:
        """
        Convert self.djvu_file to bundled format using djvmcvt.

        Returns:
            Path to bundled file
        """
        djvu_path = self.djvu_file.path

        if output_path is None:
            dirname = os.path.dirname(djvu_path)
            basename = os.path.basename(djvu_path)
            stem = os.path.splitext(basename)[0]
            output_path = os.path.join(dirname, f"{stem}_bundled.djvu")

        cmd = f"djvmcvt -b {shlex.quote(djvu_path)} {shlex.quote(output_path)}"
        self.run_cmd(cmd, "Failed to bundle DjVu file")

        if not os.path.exists(output_path):
            raise RuntimeError(f"Bundled file not created: {output_path}")

        return output_path

    @classmethod
    def convert_djvu_to_ppm(
        cls,
        djvu_path: str,
        page_num: int,
        output_path: str,
        size: str = None,  # e.g., "2480x3508" for A4 @ 300dpi
        shell: Shell = None,
        debug: bool = False,
    ) -> None:
        """Convert DJVU page to PPM using ddjvu CLI."""
        # Build command parts
        cmd_parts = [
            "ddjvu",
            "-format=ppm",
            f"-page={page_num + 1}",
        ]

        if size:
            cmd_parts.append(f"-size={size}")

        cmd_parts.extend(
            [
                shlex.quote(djvu_path),
                shlex.quote(output_path),
            ]
        )

        cmd = " ".join(cmd_parts)
        result = shell.run(cmd, text=True, debug=debug)

        if result.returncode != 0:
            raise RuntimeError(
                f"ddjvu failed (rc={result.returncode}):\n"
                f"stdout: {result.stdout}\n"
                f"stderr: {result.stderr}"
            )

    @classmethod
    def convert_ppm_to_png(cls, ppm_path: str, png_path: str) -> None:
        """Convert PPM to PNG using PIL."""
        img = Image.open(ppm_path)
        img.save(png_path, "PNG")

    @classmethod
    def render_djvu_page_cli(
        cls,
        djvu_path: str,
        page_num: int,
        output_path: str,
        size: str,  # e.g., "2480x3508" for A4 @ 300dpi
        debug: bool = False,
        shell: Shell = None,
    ) -> str:
        """Render a DJVU page to PNG using ddjvu CLI."""
        if shell is None:
            shell = Shell()

        # Create temporary PPM file
        with tempfile.NamedTemporaryFile(suffix=".ppm", delete=False) as tmp_file:
            tmp_ppm_path = tmp_file.name

        try:
            # Step 1: DJVU → PPM
            cls.convert_djvu_to_ppm(
                djvu_path=djvu_path,
                page_num=page_num,
                output_path=tmp_ppm_path,
                size=size,
                shell=shell,
                debug=debug,
            )

            # Step 2: PPM → PNG
            cls.convert_ppm_to_png(tmp_ppm_path, output_path)

            return output_path

        finally:
            # Clean up
            if os.path.exists(tmp_ppm_path):
                os.remove(tmp_ppm_path)

    def is_valid(self) -> bool:
        """Check if the bundle has no errors."""
        return len(self.errors) == 0

    def get_error_summary(self) -> str:
        """Get a formatted summary of all errors."""
        if not self.errors:
            return "No errors found"

        return f"Found {len(self.errors)} error(s):\n" + "\n".join(
            f"  - {error}" for error in self.errors
        )



"""
DjVu debug/info page.

Created on 2026-01-02

@author: wf
"""
import os
from pathlib import Path
import urllib.parse

from djvuviewer.djvu_bundle import DjVuBundle
from djvuviewer.djvu_core import DjVuPage
from djvuviewer.djvu_processor import DjVuProcessor
from ngwidgets.lod_grid import ListOfDictsGrid
from ngwidgets.progress import NiceguiProgressbar
from ngwidgets.widgets import Link
from nicegui import background_tasks, run, ui

from djvuviewer.djvu_context import DjVuContext


class DjVuDebug:
    """
    UI for displaying debug/info page for a DjVu document.
    """

    def __init__(
        self,
        solution,
        context:DjVuContext,
        page_title: str,
    ):
        """
        Initialize the DjVu debug view.

        Args:
            solution: The solution instance
            context: context with proc and actions
            page_title: pagetitle of the DjVu file
        """
        self.solution = solution
        self.context=context
        self.config = context.config
        self.webserver = self.solution.webserver
        self.progressbar = None
        self.page_title = page_title
        self.mw_image = None
        self.mw_image_new = None
        self.djvu_file = None
        self.djvu_bundle=None
        self.total_pages = 0
        self.view_lod = []
        self.lod_grid = None
        self.load_task = None
        self.timeout = 30.0  # Longer timeout for DjVu processing
        self.ui_container = None
        self.bundle_state_container=None
        self.dproc = DjVuProcessor(
            verbose=self.solution.debug, debug=self.solution.debug
        )

    def authenticated(self) -> bool:
        """
        check authentication
        """
        allow = self.solution.webserver.authenticated()
        return allow

    def load_djvu_file(self) -> bool:
        """
        Load DjVu file metadata via DjVuProcessor.

        Returns:
            bool: True if successful, False otherwise
        """
        success = False
        try:
            self.mw_image = self.solution.webserver.mw_client_base.fetch_image(
                title=self.page_title
            )
            if self.solution.webserver.mw_client_new:
                self.mw_image_new = self.solution.webserver.mw_client_new.fetch_image(
                    title=self.page_title
                )
            if self.mw_image:
                relpath = self.config.extract_and_clean_path(self.mw_image.url)
                abspath = self.config.djvu_abspath(f"/images/{relpath}")
                self.djvu_file = self.dproc.get_djvu_file(
                    abspath, config=self.config, progressbar=self.progressbar
                )
                self.djvu_bundle = DjVuBundle(self.djvu_file, config=self.config, debug=self.context.args.debug)
                success = True

        except Exception as ex:
            self.solution.handle_exception(ex)
            raise
        return success

    def get_header_html(self) -> str:
        """Helper to generate HTML summary our DjVuFile instance."""
        djvu_file=self.djvu_file
        if not djvu_file:
            markup="<div>No DjVu file information loaded.</div>"
            return markup

        format_type = "Bundled" if djvu_file.bundled else "Indirect/Indexed"

        # Safe aggregations
        total_page_size = sum((p.filesize or 0) for p in (djvu_file.pages or []))

        # Safe first page access
        first_page = djvu_file.pages[0] if djvu_file.pages else None

        dims = (
            f"{first_page.width}×{first_page.height}"
            if (first_page and first_page.width)
            else "—"
        )
        dpi = first_page.dpi if (first_page and first_page.dpi) else "—"

        tar_info = ""
        if djvu_file.tar_filesize:
            tar_info = f"<strong>Tarball:</strong><span>{djvu_file.tar_filesize:,} bytes ({djvu_file.tar_iso_date})</span>"

        # Build HTML
        html_parts = [
            f"<div style='border: 1px solid #ddd; padding: 10px; border-radius: 4px; min-width: 300px;'>",
            "<div style='display: grid; grid-template-columns: auto 1fr; gap: 4px 12px; font-size: 0.9em;'>",
            f"<strong>Path:</strong><span style='word-break: break-all;'>{djvu_file.path}</span>",
            f"<strong>Format:</strong><span>{format_type}</span>",
            f"<strong>Pages (Doc):</strong><span>{djvu_file.page_count}</span>",
            f"<strong>Pages (Dir):</strong><span>{djvu_file.dir_pages or '—'}</span>",
            f"<strong>Dimensions:</strong><span>{dims}</span>",
            f"<strong>DPI:</strong><span>{dpi}</span>",
            f"<strong>File Date:</strong><span>{djvu_file.iso_date or '—'}</span>",
            (
                f"<strong>Main Size:</strong><span>{djvu_file.filesize:,} bytes</span>"
                if djvu_file.filesize
                else ""
            ),
            f"<strong>Pages Size:</strong><span>{total_page_size:,} bytes</span>",
            tar_info,
            "</div></div>",
        ]
        markup=f"<div style='display: flex; flex-wrap: wrap; gap: 16px;'>{''.join(html_parts)}</div>"
        return markup

    def setup_djvu_info(self):
        # Generate header HTML
        header_html = self.get_header_html()

        # Header
        ui.html(header_html)

    def update_bundle_state(self):
        if not hasattr(self, 'djvu_bundle') or self.djvu_bundle is None:
            self.bundle_state_container.clear()
            with self.bundle_state_container:
                ui.label("No bundle information available")
            return

        self.bundle_state_container.clear()
        with self.bundle_state_container:
            ui.label("Bundling State").classes("text-subtitle1 mb-2")

            # Bundled status - just a disabled checkbox
            ui.checkbox("Bundled", value=self.djvu_file.bundled).props('disable')

            # Backup file - just a disabled checkbox and download link
            backup_exists = os.path.exists(self.djvu_bundle.backup_file)
            with ui.row().classes("gap-4 items-center"):
                ui.checkbox("Backup exists", value=backup_exists).props('disable')

                if backup_exists:
                    backup_rel_path = os.path.relpath(
                        self.djvu_bundle.backup_file,
                        self.config.backup_path
                    )
                    download_url = f"{self.config.url_prefix}/backups/{backup_rel_path}"
                    ui.link(f"⬇️{backup_rel_path}", download_url).classes("text-primary")

            with ui.expansion('Bundling script', icon='code'):
                # Script
                script = self.djvu_bundle.generate_bundling_script()
                ui.code(script, language='bash').classes('w-full text-xs')


    def create_page_record(
        self, djvu_path: str, page: DjVuPage
    ) -> dict:
        """Helper to create a single dictionary record for the LOD."""
        filename_stem = Path(djvu_path).name

        record = {
            "#": page.page_index,
            "Page": page.page_index,
            "Filename": page.path or "—",
            "Valid": "✅" if page.valid else "❌",
            "Dimensions": (
                f"{page.width}×{page.height}" if (page.width and page.height) else "—"
            ),
            "DPI": page.dpi or "—",
            "Size": f"{page.filesize:,}" if page.filesize else "—",
            "Error": page.error_msg or "",
        }

        # Add Links if config exists
        if hasattr(self, "config") and hasattr(self.config, "url_prefix"):
            base_url = f"{self.config.url_prefix}/djvu"

            # View Link
            if self.mw_image_new.description_url:
                backlink = (
                    f"&backlink={urllib.parse.quote(self.mw_image_new.description_url)}"
                )
            view_url = f"{base_url}/{filename_stem}?page={page.page_index}{backlink}"
            record["view"] = Link.create(url=view_url, text="view")

            # PNG Download Link
            # Logic assumes content is served under content/{stem}/{png_file}
            stem_only = Path(filename_stem).stem
            png_url = f"{base_url}/content/{stem_only}/{page.png_file}"
            record["png"] = Link.create(url=png_url, text="png")

        return record

    def get_view_lod(self) -> list:
        """
        Convert page records into a List of Dicts by iterating over abstract sources.
        """
        view_lod = []
        if not self.djvu_file:
            return

        for page in self.djvu_file.pages:
            record = self.create_page_record(self.djvu_file.path, page)
            view_lod.append(record)
            self.total_pages += 1

        return view_lod


    async def load_debug_info(self):
        """Load DjVu file metadata and display it."""
        try:
            # Load file metadata (blocking IO)
            await run.io_bound(self.load_djvu_file)
            # Convert pages to view format
            self.view_lod = self.get_view_lod()

            # Clear and update UI
            self.content_row.clear()
            # side by side
            with self.card_row:
                with ui.splitter() as splitter:
                    with splitter.before:
                        self.setup_djvu_info()
                    with splitter.after:
                        with ui.element("div").classes("w-full") as self.bundle_state_container:
                            self.update_bundle_state()

            with self.content_row:
                # Grid
                self.lod_grid = ListOfDictsGrid()
                self.lod_grid.load_lod(self.view_lod)

            if self.lod_grid:
                self.lod_grid.sizeColumnsToFit()

            with self.solution.container:
                self.content_row.update()

        except Exception as ex:
            self.solution.handle_exception(ex)
            self.content_row.clear()
            with self.content_row:
                ui.notify(f"Error loading DjVu file: {str(ex)}", type="negative")
                ui.label(f"Failed to load: {self.page_title}").classes("text-negative")

    def reload_debug_info(self):
        """Create background task to reload debug info."""
        self.load_task = background_tasks.create(self.load_debug_info())

    async def bundle(self):
        """
        run the bundle activities in background
        """
        try:
            if os.path.exists(self.djvu_bundle.backup_file):
                with self.content_row:
                    ui.notify(f"{self.djvu_bundle.backup_file} already exists")
            else:
                self.djvu_bundle.create_backup_zip()
            self.update_bundling_state()

        except Exception as ex:
            self.solution.handle_exception(ex)


    def on_bundle(self):
        """
        handle bundle click
        """
        with self.content_row:
            self.bundle_task=background_tasks.create(self.bundle())

    def on_refresh(self):
        """Handle refresh button click."""

        def cancel_running():
            if self.load_task:
                self.load_task.cancel()

        # Show loading spinner
        self.content_row.clear()
        with self.content_row:
            ui.spinner()
        self.content_row.update()

        # Cancel any running task
        cancel_running()

        # Set timeout
        ui.timer(self.timeout, lambda: cancel_running(), once=True)

        # Reload
        self.reload_debug_info()

    def setup_ui(self):
        """Set up the user interface components for the DjVu debug page."""
        self.ui_container = self.solution.container

        # Header with refresh button
        with ui.row() as self.header_row:
            ui.label("DjVu Debug").classes("text-h6")
            self.refresh_button = ui.button(
                icon="refresh",
                on_click=self.on_refresh,
            ).tooltip("Refresh debug info")
            self.bundle_button = ui.button(
                icon="archive",
                on_click=self.on_bundle,
            ).tooltip("bundle the shown DjVu file")
            self.bundle_button.enabled = self.authenticated()
            self.progressbar = NiceguiProgressbar(
                total=1,  # Will be updated by get_djvu_file
                desc="Loading DjVu",
                unit="pages",
            )
        # side by side cards for bundle infos left: djvu right: state
        self.card_row=ui.row().classes("w-full")
        # Content row for all content
        self.content_row = ui.row()

        # Initial load
        self.reload_debug_info()



"""
Created on 2025-02-25

@author: wf
"""

import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

import numpy
from basemkit.yamlable import lod_storable

from djvuviewer.tarball import Tarball


@lod_storable
class DjVuPage:
    """Represents a single djvu page"""

    path: str
    page_index: int
    valid: bool = False
    iso_date: Optional[str] = None
    filesize: Optional[int] = None
    width: Optional[int] = None
    height: Optional[int] = None
    dpi: Optional[int] = None
    djvu_path: Optional[str] = None
    page_key: Optional[str] = None
    error_msg: Optional[str] = None

    def __post_init__(self):
        """Post-initialization logic for DjVuPage."""
        if self.page_key is None:
            # we expect no more than 9999 pages per document in the genwiki context that is proven
            self.page_key = f"{self.djvu_path}#{self.page_index:04d}"
        pass

    @property
    def png_file(self) -> str:
        """
        Returns the PNG file name derived from the DjVu file path and page index.
        """
        prefix = os.path.splitext(os.path.basename(self.djvu_path))[0]
        png_file = f"{prefix}_page_{self.page_index:04d}.png"
        return png_file

    @classmethod
    def get_sample(cls):
        """Returns a sample DjVuPage instance for testing."""
        sample_page = cls(
            path="s_455_0001.djvu",
            page_index=1,
            valid=False,
            iso_date="2009-06-02T07:17:55+00:00",
            filesize=66327,
            width=2829,
            height=4194,
            dpi=216,
            djvu_path="b/b3/AB1951-Suenninghausen.djvu",
            page_key="b/b3/AB1951-Suenninghausen.djvu#0001",
            error_msg="-sample error message-",
        )
        return sample_page


@dataclass
class DjVu:
    """Represents a DjVu main file e.g. bundled or indexed"""

    path: str
    page_count: int
    bundled: bool = False
    iso_date: Optional[str] = None
    filesize: Optional[int] = None
    tar_filesize: Optional[int] = None
    tar_iso_date: Optional[str] = None
    dir_pages: Optional[int] = None

    @classmethod
    def get_sample(cls):
        """Returns a sample DjVu instance for testing."""
        sample_djvu = cls(
            path="images/b/b3/AB1951-Suenninghausen.djvu",
            iso_date="2009-06-02",
            filesize=85,
            tar_filesize=0,
            tar_iso_date="2026-01-02",
            page_count=4,
            dir_pages=5,
            bundled=False,
        )
        return sample_djvu


@lod_storable
class DjVuFile(DjVu):
    """Represents a DjVu main file e.g. bundled or indexed"""

    pages: List[DjVuPage] = field(default_factory=list)

    def get_page_by_page_index(self, page_index: int) -> Optional[DjVuPage]:
        """
        Retrieve a page by its page index.

        Args:
            page_index (int): The index of the page to retrieve.

        Returns:
            Optional[DjVuPage]: The requested DjVuPage if found, otherwise None.
        """
        for page in self.pages:
            if page.page_index == page_index:
                return page
        return None

    @classmethod
    def from_tarball(cls, tarball_path: Path, yaml_filename: str) -> "DjVuFile":
        """
        (re)-instantiate me from a YAML serialization in a tarball
        """
        yaml_data = Tarball.read_from_tar(tarball_path, yaml_filename).decode("utf-8")
        djvu_file = cls.from_yaml(yaml_data)
        return djvu_file


@dataclass
class DjVuViewPage:
    file: DjVuFile
    page: DjVuPage
    base_path: str

    @property
    def content_path(self) -> str:
        """Path for content retrieval"""
        return f"{Path(self.base_path).stem}/{self.page.png_file}"

    @property
    def image_url(self) -> str:
        """URL path for HTML display"""
        return f"/djvu/content/{self.content_path}"


@dataclass
class DjVuImage(DjVuPage):
    _buffer: Optional[numpy.ndarray] = None



"""
Created on 2026-01-01

@author: wf
"""

import os
import pathlib
import re
import urllib.parse
from enum import Enum
from typing import Optional

from basemkit.yamlable import lod_storable


class PngMode(Enum):
    """PNG generation mode"""

    CLI = "cli"  # Use ddjvu command-line tool
    PIL = "pil"  # Use PIL with rendered buffer


@lod_storable
class DjVuConfig:
    """
    configuration for DjVu Viewer and Converter
    """

    # singleton
    _instance: Optional["DjVuConfig"] = None

    tarball_path: Optional[str] = None
    images_path: Optional[str] = None
    db_path: Optional[str] = None
    queries_path: Optional[str] = None
    backup_path: Optional[str] = None  # Path for bundle backups
    container_name: Optional[str] = None  # MediaWiki container name for maintenance
    base_url: Optional[str] = "https://wiki.genealogy.net/"
    new_url: Optional[str] = None
    url_prefix: Optional[str] = (
        ""  # URL prefix for proxied deployments (e.g., "/djvu-viewer")
    )

    def __post_init__(self):
        """
        make sure we set defaults
        """
        examples_path = DjVuConfig.get_examples_path()
        if self.queries_path is None:
            self.queries_path = os.path.join(examples_path, "djvu_queries.yaml")
        if self.tarball_path is None:
            self.tarball_path = os.path.join(examples_path, "djvu_images")
        if self.images_path is None:
            self.images_path = os.path.join(examples_path, "images")
        if self.db_path is None:
            self.db_path = os.path.join(examples_path, "djvu_data.db")
        if self.backup_path is None:
            self.backup_path = os.path.join(examples_path, "backup")
        if self.container_name is None:
            self.container_name = "genwiki39-mw"

    def djvu_relpath(self, path: str) -> str:
        """Convert path to wiki image-relative format by removing './' and '/images/'."""

        # Look for 'images/' anywhere in the path and extract everything after it
        match = re.search(r"images/(.*)", path)

        if match:
            # Extract the part after 'images/' and prepend '/'
            cleaned_path = "/" + match.group(1)
        else:
            # No 'images/' found - just handle './' prefix
            if path.startswith("./"):
                cleaned_path = "/" + path[2:]
            else:
                cleaned_path = path

        # Remove duplicate slashes
        cleaned_path = re.sub(r"/+", "/", cleaned_path)

        return cleaned_path

    def wiki_fileurl(
        self, filename: str, new: bool = False, quoted: bool = False
    ) -> str:
        """get the wiki file url for the given filename"""
        url = self.new_url if new else self.base_url
        # wiki_url = f"{self.base_url}/File:{filename}"
        wiki_url = urllib.parse.urljoin(url, f"index.php?title=File:{filename}")
        if quoted:
            wiki_url = urllib.parse.quote(wiki_url)
        return wiki_url

    def djvu_abspath(self, path: str) -> str:
        """Get absolute DjVu path by prepending images_path to relative path."""
        djvu_path = self.images_path + self.djvu_relpath(path)
        return djvu_path

    def extract_and_clean_path(self, url: str) -> str:
        """
        URL decode, extract path from /images, and remove duplicate slashes.

        Args:
            url (str): The URL to process

        Returns:
            str: The cleaned path starting from /images
        """
        # URL decode
        decoded_url = urllib.parse.unquote(url)
        relpath = self.djvu_relpath(decoded_url)
        return relpath

    @classmethod
    def get_config_file_path(cls) -> str:
        """
        Returns the standard location for the config file: $HOME/.djvuviewer/config.yaml
        """
        home = pathlib.Path.home()
        config_dir = home / ".djvuviewer"
        config_dir.mkdir(parents=True, exist_ok=True)
        return str(config_dir / "config.yaml")

    @classmethod
    def get_instance(cls, test: bool = False) -> "DjVuConfig":
        """
        get my instance
        """
        if cls._instance is None:
            config_path = cls.get_config_file_path()
            if os.path.exists(config_path) and not test:
                # load_from_yaml_file is provided by the @lod_storable decorator
                instance = cls.load_from_yaml_file(config_path)
            else:
                # Return default instance if no config file found
                instance = cls()
            cls._instance = instance
        return cls._instance

    @classmethod
    def get_examples_path(cls) -> str:
        # the root directory (default: examples)
        path = os.path.join(os.path.dirname(__file__), "../djvuviewer_examples")
        path = os.path.abspath(path)
        return path



"""
Created on 2026-01-03

@author: wf
"""

import datetime
import logging
import os
from dataclasses import dataclass, field
from typing import Optional, Tuple

import djvu
from basemkit.profiler import Profiler

from djvuviewer.djvu_core import DjVuImage


@dataclass
class ImageJob:
    """
    Represents a processed DjVu page,
    optionally including document, page, page job, and image data.
    """

    djvu_path: str  # fully qualifying path of the container DjVu document
    document: djvu.decode.Document
    page: djvu.decode.Page
    page_index: int  # page_index to track position
    relurl: str  # relurl for context
    pagejob: Optional[djvu.decode.PageJob] = field(default=None)
    image: Optional[DjVuImage] = field(default=None)
    # flags
    verbose: bool = False
    debug: bool = False
    # fields for database records
    iso_date: Optional[str] = field(default=None)
    filesize: Optional[int] = field(default=None)
    error: Optional[Exception] = field(default=None)

    def __post_init__(self):
        """Initialize profiler if not provided"""
        self.profiler = Profiler(
            f"Image Job {self.relurl}#{self.page_index:04d}",
            profile=self.verbose or self.debug,
        )
        self.profiler.start()

    def log(self, msg):
        if self.verbose or self.debug:
            self.profiler.time(" " + msg)

    def get_size(self) -> Tuple[int, int]:
        """Get the width and height of the page if pagejob is available"""
        if self.pagejob:
            return self.pagejob.size
        return (0, 0)

    @staticmethod
    def get_fileinfo(filepath: str):
        filesize = None
        iso_date = None
        if os.path.exists(filepath):
            # Set file size in bytes
            filesize = os.path.getsize(filepath)

            # Get file modification time and convert to UTC ISO format with second precision
            mtime = os.path.getmtime(filepath)
            datetime_obj = datetime.datetime.fromtimestamp(
                mtime, tz=datetime.timezone.utc
            )
            iso_date = datetime_obj.isoformat(timespec="seconds")
        return iso_date, filesize

    def set_fileinfo(self, filepath: str):
        """
        Set filesize and ISO date with sec prec for
        the given filepath

        Args:
            filepath (str): Path to the file
        """
        self.iso_date, self.filesize = self.get_fileinfo(filepath)

    @staticmethod
    def get_prefix(relurl: str):
        prefix = os.path.splitext(os.path.basename(relurl))[0]
        return prefix

    @staticmethod
    def get_relative_image_path(relurl: str):
        image_rel_dir = os.path.dirname(relurl)
        # Ensure image_rel_dir is treated as a relative path
        if image_rel_dir.startswith(
            os.sep
        ):  # os.sep is '/' on Unix and '\\' on Windows
            image_rel_dir = image_rel_dir.lstrip(os.sep)
        return image_rel_dir

    @property
    def prefix(self) -> str:
        prefix = ImageJob.get_prefix(relurl=self.relurl)
        return prefix

    @property
    def filename(self) -> str:
        try:
            # Attempt to safely decode the file name
            filename = self.page.file.name.encode("utf-8", errors="replace").decode(
                "utf-8"
            )
        except Exception as e:
            if self.debug:
                logging.warn(
                    f"Failed to decode filename for page {self.page_index}: {e}"
                )
            filename = f"page_{self.page_index:04d}.djvu"
        return filename

    @property
    def dirname(self) -> str:
        dirname = os.path.dirname(self.djvu_path)
        return dirname

    @property
    def filepath(self) -> str:
        filepath = os.path.join(self.dirname, self.filename)
        return filepath



"""
Created on 2025-03-04

@author: wf
"""

import io

from PIL import Image


class ImageConverter:
    """
    Converter for image formats with DPI and scale handling
    """

    def __init__(self, png_data: bytes, dpi: int):
        """
        Initialize the image converter

        Args:
            png_data: PNG image data as bytes
            dpi: Original DPI of the image
        """
        self.png_data = png_data
        self.dpi = dpi

    def convert_to_jpg(
        self, target_dpi: int = None, scale: float = None, quality: int = 85
    ) -> bytes:
        """
        Convert PNG to JPG with optional scaling and quality settings to reduce file size/loading time

        Args:
            target_dpi: Target DPI for scaling. If provided, calculates scale based on original DPI.
            scale: Direct scale factor (0.0-1.0). Overrides target_dpi if both are provided.
            quality: JPEG quality (1-100), lower values = smaller file size, faster loading

        Returns:
            JPG image data as bytes
        """
        # Open image from bytes
        img = Image.open(io.BytesIO(self.png_data))

        # Determine scaling factor
        if scale is not None:
            # Use direct scale factor
            scaling_factor = scale
        elif target_dpi is not None and target_dpi < self.dpi:
            # Calculate scale based on DPI
            scaling_factor = target_dpi / self.dpi
        else:
            # No scaling
            scaling_factor = 1.0

        # Apply scaling if needed
        if scaling_factor < 1.0:
            new_width = int(img.width * scaling_factor)
            new_height = int(img.height * scaling_factor)
            img = img.resize((new_width, new_height), Image.Resampling.LANCZOS)

        # Convert to JPEG with specified quality
        output = io.BytesIO()
        img.convert("RGB").save(
            output,
            format="JPEG",
            quality=quality,
            optimize=True,  # Enable JPEG optimization
        )

        # Get the bytes from the BytesIO object and assign to variable
        jpg_bytes = output.getvalue()

        # Return the bytes
        return jpg_bytes



"""
Created on 2026-01-02

@author: wf
"""

import datetime
from dataclasses import asdict

from basemkit.basetest import Basetest

from djvuviewer.wiki_images import MediaWikiImages


class TestMediaWikiImages(Basetest):
    """
    Test MediaWiki Images handling
    """

    def setUp(self, debug=False, profile=True):
        """
        setUp test environment
        """
        Basetest.setUp(self, debug=debug, profile=profile)

        self.mwi = MediaWikiImages(
            api_url="https://genwiki.genealogy.net/api.php",
            mime_types=("image/vnd.djvu", "image/x-djvu"),
            aiprop=("url", "mime", "size", "timestamp", "user"),
            timeout=20,
        )

    def testFetchSingleImage(self):
        """
        test fetching a single image as a dataclass object
        """
        title = "Datei:AB1938_Heessen-Geschi.djvu"
        image = self.mwi.fetch_image(title)
        image_dict = asdict(image)
        if self.debug:
            print(image)
            print(image_dict)
        expected = {
            "name": "Datei:AB1938 Heessen-Geschi.djvu",
            "url": "https://wiki.genealogy.net/images//0/0c/AB1938_Heessen-Geschi.djvu",
            "mime": "image/vnd.djvu",
            "size": 161771,
            "user": "KlausErdmann",
            "timestamp": datetime.datetime(2008, 5, 17, 10, 0, 3),
            "description_url": "https://wiki.genealogy.net/Datei:AB1938_Heessen-Geschi.djvu",
            "height": 2689,
            "width": 2095,
        }
        self.assertEqual(image_dict, expected)

    def testFetchAllImages(self):
        """
        test fetching all images
        """
        limit = 3
        images = self.mwi.fetch_allimages(limit=limit)
        for img in images:
            if self.debug:
                print(img)
        self.assertEqual(len(images), limit)



"""
Created on 2025-02-24

@author: wf
"""

import argparse
import glob
import json
import os
from argparse import Namespace
from typing import List, Optional

from basemkit.basetest import Basetest

from djvuviewer.djvu_bundle import DjVuBundle
from djvuviewer.djvu_cmd import DjVuCmd
from djvuviewer.djvu_config import DjVuConfig
from djvuviewer.djvu_core import DjVuFile, DjVuImage
from djvuviewer.djvu_image import ImageJob
from djvuviewer.djvu_manager import DjVuManager
from djvuviewer.djvu_processor import DjVuProcessor
from djvuviewer.download import Download


class TestDjVu(Basetest):
    """
    Test djvu handling
    """

    def setUp(self, debug=True, profile=True):
        """
        setUp test environment
        """
        Basetest.setUp(self, debug=debug, profile=profile)
        # Define base directory
        base_dir = os.path.expanduser("/tmp/djvu")

        # Set up subdirectories
        self.output_dir = os.path.join(base_dir, "test_pngs")
        self.db_path = os.path.join(base_dir, "test_db", "genwiki_images.db")
        self.backup_path = os.path.join(base_dir, "backup")
        # Create all necessary directories
        os.makedirs(self.output_dir, exist_ok=True)
        os.makedirs(os.path.dirname(self.db_path), exist_ok=True)
        os.makedirs(self.backup_path, exist_ok=True)
        self.local = os.path.exists(DjVuConfig.get_config_file_path())
        # set to True to emulate CI mode
        force_test = False
        if force_test:
            self.local = False
        self.config = DjVuConfig.get_instance(test=force_test)
        self.config.backup_path = self.backup_path
        self.limit = 50 if not self.local else 50  # 10000000
        self.test_tuples = [
            ("/images/c/c7/AB1938_Kreis-Beckum_Inhaltsverz.djvu", 3, False),
            ("/images/c/ce/Plauen-AB-1938.djvu", 2, True),
            ("/images/f/ff/AB1932-Ramrath.djvu", 2, True),
        ]
        self.test_bundles = None
        self.test_tuples_2024 = [
            ("/images/2/2f/Sorau-AB-1913.djvu", 255, False),
            ("/images/9/96/vz1890-neuenhausen-zb04.djvu", 3, True),
            ("/images/0/08/Deutsches-Kirchliches-AB-1927.djvu", 1188, False),
            ("/images/a/a1/Treuen-Vogtland-AB-1905.djvu", 38, False),
        ]
        self.dproc = DjVuProcessor()

    def get_djvu_test_bundles(self) -> List[DjVuFile]:
        """
        get the djvu test bundles by derived from unbundled/indexed DjVu test tuples
        """
        if self.test_bundles is None:
            self.test_bundles = []

            for relurl, _elen, expected_bundled in self.test_tuples:
                if not expected_bundled:
                    djvu_path = self.get_djvu(relurl)
                    rel_path = self.config.djvu_relpath(djvu_path)
                    if self.debug:
                        print(f"getting DjVuFile for {rel_path}")
                    djvu_file = self.dproc.get_djvu_file(djvu_path, config=self.config)
                    djvu_bundle = DjVuBundle(djvu_file, config=self.config)
                    self.test_bundles.append(djvu_bundle)
        return self.test_bundles

    def get_args(self, command: str) -> argparse.Namespace:
        """
        get CLI arguments for testing
        """
        args = argparse.Namespace(
            command=command,
            db_path=self.db_path,
            images_path=self.config.images_path,
            backup_path=self.backup_path,
            container_name="genwiki39-mw",
            limit=self.limit,
            url=None,
            sort="asc",
            force=False,
            output_path=self.output_dir,
            parallel=False,
            batch_size=100,
            limit_gb=16,
            max_errors=1,
            max_workers=None,
            debug=self.debug,
            pngmode="pil",
            verbose=self.debug,
            quiet=False,
            about=False,
            serial=False,
        )
        return args

    def check_command(
        self, command: str, expected_errors: int = 0, args: Namespace = None
    ):
        """
        check the given command
        """
        if args is None:
            args = self.get_args(
                command=command,
            )
        djvu_cmd = DjVuCmd(args=args)
        djvu_cmd.handle_args(args)
        error_count = len(djvu_cmd.actions.errors)
        if self.debug and error_count > expected_errors:
            print(djvu_cmd.actions.errors)
        self.assertTrue(error_count <= expected_errors)

    def test_djvu_dump(self):
        """
        test djvu_dump
        """
        for djvu_bundle in self.get_djvu_test_bundles():
            if self.debug:
                print(djvu_bundle.djvu_file.to_yaml())
            log = djvu_bundle.djvu_dump()
            if self.debug:
                print(log)
            part_filenames = djvu_bundle.get_part_filenames()
            if self.debug:
                for i, filename in enumerate(part_filenames):
                    print(f"{i}:{filename}")

    def show_fileinfo(self, path: str) -> int:
        """
        show info for a file
        """
        iso_date, filesize = ImageJob.get_fileinfo(path)
        if self.debug:
            print(f"{path} ({filesize}) {iso_date}")
        return filesize

    def test_bundle(self):
        """
        Test zipping and bundling with size ratio validation.
        """
        for djvu_bundle in self.get_djvu_test_bundles():
            # Create backup zip
            zip_path = djvu_bundle.create_backup_zip()
            zip_filesize = self.show_fileinfo(zip_path)

            # Convert to bundled format
            bundled_path = djvu_bundle.convert_to_bundled()
            bundle_filesize = self.show_fileinfo(bundled_path)

            # Calculate and validate size ratio
            if zip_filesize > 0:
                size_ratio = zip_filesize / bundle_filesize

                # Assert bundled file is slightly smaller than zip
                assert 1.0 < size_ratio < 1.2, (
                    f"Unexpected size ratio: {size_ratio:.2f}. "
                    f"Bundle: {bundle_filesize} bytes, "
                    f"Zip: {zip_filesize} bytes. "
                    f"Files: {zip_path}, {bundled_path}"
                )
            else:
                self.fail(f"Zero-size zip file: {zip_path}")

    def test_config(self):
        """
        test the configuration
        """
        if self.debug:
            print(self.config)
        self.assertEqual(self.local, not self.inPublicCI())

    def test_relpath(self):
        """
        test relpath
        """
        # Test cases: (input, expected_output, description)
        test_cases = [
            # Relative paths that should be transformed
            (
                "./images/c/ce/Plauen-AB-1938.djvu",
                "/c/ce/Plauen-AB-1938.djvu",
                "Relative path with ./images/",
            ),
            (
                "./c/ce/Plauen-AB-1938.djvu",
                "/c/ce/Plauen-AB-1938.djvu",
                "Relative path with ./",
            ),
            # Absolute paths get everything before and including /images/ removed
            (
                "/Users/wf/hd/genwiki_gruff/images/c/ce/Plauen-AB-1938.djvu",
                "/c/ce/Plauen-AB-1938.djvu",
                "Absolute path with /images/ in middle",
            ),
            # Path starting with /images/ should have it removed
            (
                "/images/c/ce/Plauen-AB-1938.djvu",
                "/c/ce/Plauen-AB-1938.djvu",
                "Path starting with /images/",
            ),
            # corner case images/ prefix needs to be handled
            (
                "images/c/ce/Plauen-AB-1938.djvu",
                "/c/ce/Plauen-AB-1938.djvu",
                "Path starting with images/ (no dot or slash)",
            ),
            # duplicate slashes
            (
                "/images//f/ff/AB1932-Ramrath.djvu",
                "/f/ff/AB1932-Ramrath.djvu",
                "duplicate slashes",
            ),
        ]

        for input_path, expected, description in test_cases:
            result = self.config.djvu_relpath(input_path)
            if self.debug:
                print(f"\n{description}")
                print(f"  Input:    {input_path}")
                print(f"  Expected: {expected}")
                print(f"  Got:      {result}")
                print(f"  Match:    {'✓' if result == expected else '✗'}")

            self.assertEqual(
                expected, result, f"Failed for {description}: input='{input_path}'"
            )

    def test_001_init_db(self):
        """
        check init database command first
        """
        self.check_command("initdb")

    def get_djvu(self, relurl, with_download: bool = False):
        """
        get the djvu file for the relative url
        """
        djvu_path = self.config.djvu_abspath(relurl)
        url = self.config.base_url + relurl
        if not self.local and with_download:
            try:
                Download.download(url, djvu_path)
            except Exception as _ex:
                print(f"invalid {djvu_path}")
                return None
        self.assertTrue(os.path.isfile(djvu_path), djvu_path)
        return djvu_path

    def test_djvu_files(self):
        """
        test the djvu file operations
        """

        for relurl, elen, expected_bundled in self.test_tuples:
            djvu_path = self.get_djvu(relurl)
            rel_path = self.config.djvu_relpath(djvu_path)
            if self.debug:
                print(f"getting DjVuFile for {rel_path}")
            djvu_file = self.dproc.get_djvu_file(djvu_path, config=self.config)
            if self.debug:
                print(djvu_file)
                print(djvu_file.to_yaml())
            self.assertEqual(expected_bundled, djvu_file.bundled)
            self.assertEqual(elen, djvu_file.page_count)

    def test_djvu_images(self):
        """
        test the DjVu image generation
        """
        for relurl, elen, expected_bundled in self.test_tuples:
            djvu_path = self.get_djvu(relurl)
            dproc = DjVuProcessor(verbose=self.debug, debug=self.debug)

            if self.debug:
                print(f"processing images for {relurl}")

            count = 0
            # iterate over the generator
            for image_job in dproc.process(
                djvu_path, relurl, save_png=False, output_path=self.output_dir
            ):
                count += 1
                self.assertIsNone(image_job.error)
                image = image_job.image

                # Check ImageJob integrity
                self.assertIsNotNone(image, f"Image should be present for page {count}")
                self.assertIsInstance(image, DjVuImage)

                # Check DjVuImage properties
                self.assertIsNotNone(image._buffer, "Image buffer should not be None")
                self.assertGreater(image.width, 0, "Width should be positive")
                self.assertGreater(image.height, 0, "Height should be positive")
                self.assertGreaterEqual(image.page_index, count, "Page index mismatch")
                if self.debug:
                    print(image.to_yaml())
            self.assertGreaterEqual(
                elen, count, f"Expected {elen} images but got {count}"
            )

    def test_queries(self):
        """
        test all queries
        """
        query_params = {
            "all_pages": {"limit": 50},
            "pages_of_djvu": {"djvu_path": "/images/a/a1/Treuen-Vogtland-AB-1905.djvu"},
        }
        djvm = DjVuManager(config=self.config)
        djvm.sql_db.debug = self.debug
        # Get all available queries from the MultiLanguageQueryManager
        for query_name in djvm.mlqm.query_names:
            if self.debug:
                print(query_name)
            param_dict = query_params.get(query_name, {})
            if param_dict:
                pass
            lod = djvm.query(query_name, param_dict=param_dict)
            if self.debug:
                print(f"{len(lod)} records")

    def test_update_database(self):
        """
        test updating the database
        """
        for relurl, _elen, _expected_bundled in self.test_tuples:
            args = self.get_args("dbupdate")
            args.url = relurl
            self.check_command("dbupdate", args=args)

    def test_all_djvu(self):
        """
        test all djvu pages
        """
        expected_errors = 0 if self.local else 49
        self.check_command("catalog", expected_errors)

    def check_tarball(self, tar_file: str, relurl: Optional[str] = None):
        """
        Test helper: Verify tarball and assert validity.

        Args:
            tar_file: Path to the tar file to validate
            relurl: Optional relative URL for error context
        """
        bundle = DjVuBundle.from_tarball(tar_file, with_check=False)

        # If relurl provided, re-check with context (bundle.check_tarball is idempotent)
        if relurl:
            bundle.check_tarball(tar_file, relurl)

        self.assertTrue(
            bundle.is_valid(),
            f"Tarball validation failed for {relurl or tar_file}:\n{bundle.get_error_summary()}",
        )

    def test_convert(self):
        """
        Test the conversion with different PNG modes.
        """
        png_modes = ["pil", "cli"]

        for relurl, _elen, _expected_bundled in self.test_tuples:
            base_name = os.path.splitext(os.path.basename(relurl))[0]

            for pngmode in png_modes:
                with self.subTest(relurl=relurl, pngmode=pngmode):
                    args = self.get_args("convert")
                    args.url = relurl
                    args.force = True
                    args.pngmode = pngmode
                    self.check_command("convert", args=args)

                    # Verify tar file was created and contains expected content
                    tar_file = os.path.join(self.output_dir, f"{base_name}.tar")
                    self.check_tarball(tar_file, relurl)

    def test_issue49(self):
        """
        Test loading DjVu file with python-djvu and storing relevant metadata.
        """
        for url, page_count in [
            ("/images/9/96/vz1890-neuenhausen-zb04.djvu", 3),
            # ("f/fc/Siegkreis-AB-1905-06_Honnef.djvu", 35),
            # ("./images/9/96/Elberfeld-AB-1896-97-Stadtplan.djvu", 1),
            # ("./images/0/08/Deutsches-Kirchliches-AB-1927.djvu", 1188),
        ]:
            with self.subTest(url=url, expected_pages=page_count):
                if not self.local and page_count > 1:
                    return
                relurl = url
                djvu_path = self.get_djvu(relurl)
                dproc = DjVuProcessor(tar=False, debug=self.debug, verbose=self.debug)
                if self.debug:
                    print(f"processing {relurl}")
                # for document, page in dproc.yield_pages(djvu_path):
                #    pass
                count = 0
                for _image_job in dproc.process_parallel(
                    djvu_path, relurl=relurl, save_png=True, output_path=self.output_dir
                ):
                    count += 1

                if self.debug:
                    print(f"Processed {count} pages in {self.output_dir}")
                    base_name = os.path.splitext(os.path.basename(relurl))[0]
                pattern = os.path.join(self.output_dir, f"{base_name}_page_*.png")
                png_files = glob.glob(pattern)

                self.assertEqual(
                    len(png_files),
                    page_count,
                    f"Expected {page_count} PNG files matching pattern '{pattern}', but found {len(png_files)}",
                )

    def testDjVuManager(self):
        """
        test the DjVu Manager
        """
        dvm = DjVuManager(config=self.config)
        lod = dvm.query("total")
        if self.debug:
            print(json.dumps(lod, indent=2))
        if self.local:
            self.assertEqual(lod, [{"files": 4288, "pages": 1006670}])
        else:
            self.assertEqual(lod, [{"files": 1, "pages": 4}])
